"""
Module with an implementation of the Gayton Hat test function.

The Gayton Hat function is a two-dimensional limit-state function used
in [1] as a test function for reliability analysis algorithms.

References
----------

1. B. Echard, N. Gayton, M. Lemaire, and N. Relun, “A combined Importance
   Sampling and Kriging reliability method for small failure probabilities
   with time-demanding numerical models”, Reliability Engineering
   and System Safety, vol. 111, pp. 232-240, 2013.
   DOI: 10.1016/j.ress.2012.10.008
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["GaytonHat"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Echard2013": {
        "function_id": "GaytonHat",
        "description": (
            "Input model for the Gayton Hat function "
            "from Echard et al. (2013)"
        ),
        "marginals": [
            {
                "name": "U1",
                "distribution": "normal",
                "parameters": [0, 1],
                "description": None,
            },
            {
                "name": "U2",
                "distribution": "normal",
                "parameters": [0, 1],
                "description": None,
            },
        ],
        "copulas": None,
    },
}


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the Gayton Hat function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        A two-Dimensional input values given by an N-by-2 array
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = 0.5 * (xx[:, 0] - 2) ** 2 - 1.5 * (xx[:, 1] - 5) ** 3 - 3

    return yy


class GaytonHat(UQTestFunFixDimABC):
    """A concrete implementation of the Gayton Hat test function."""

    _tags = ["reliability"]
    _description = (
        "Two-Dimensional Gayton Hat function from Echard et al. (2013)"
    )
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None
    _default_input_dimension = 2

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with an implementation the one-dimensional damped cosine function.


References
----------

1. Jeremy Oakley and Anthony O'Hagan, "Bayesian inference for the uncertainty
   distribution of computer model outputs," Biometrika , Vol. 89, No. 4,
   p. 769-784, 2002.
   DOI: 10.1093/biomet/89.4.769
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["DampedCosine"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Santner2018": {
        "function_id": "DampedCosine",
        "description": (
            "Input model for the one-dimensional damped cosine "
            "from Santner et al. (2018)"
        ),
        "marginals": [
            {
                "name": "x",
                "distribution": "uniform",
                "parameters": [0.0, 1.0],
                "description": None,
            },
        ],
        "copulas": None,
    }
}


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the 1D damped cosine function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        1-Dimensional input values given by an N-by-1 array
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the 1D Oakley-O'Hagan function evaluated
        on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = np.exp(-1.4 * xx[:, 0]) * np.cos(3.5 * np.pi * xx[:, 0])

    return yy


class DampedCosine(UQTestFunFixDimABC):
    """An implementation of the 1D damped cosine from Santner et al. (2018)."""

    _tags = ["metamodeling"]
    _description = "One-dimensional damped cosine from Santner et al. (2018)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None
    _default_input_dimension = 1

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with an implementation of the Ishigami function.

The Ishigami test function is a three-dimensional scalar-valued function.
The function was first introduced in [1] in the context of sensitivity
analysis, and has been revisited many times in the same context
(see, for instances, [2], [3], [4]).

References
----------

1. T. Ishigami and T. Homma, “An importance quantification technique in
   uncertainty analysis for computer models,” in [1990] Proceedings.
   First International Symposium on Uncertainty Modeling and Analysis,
   College Park, MD, USA, 1991, pp. 398–403.
   DOI: 10.1109/ISUMA.1990.151285.
2. I. M. Sobol’ and Y. L. Levitan, “On the use of variance reducing multipliers
   in Monte Carlo computations of a global sensitivity index,”
   Computer Physics Communications, vol. 117, no. 1, pp. 52–61, 1999.
   DOI: 10.1016/S0010-4655(98)00156-8
3. B. Sudret, “Global sensitivity analysis using polynomial chaos expansions,”
   Reliability Engineering & System Safety, vol. 93, no. 7, pp. 964–979, 2008.
   DOI: 10.1016/j.ress.2007.04.002.
4. A. Marrel, B. Iooss, B. Laurent, and O. Roustant, "Calculations of
   Sobol indices for the Gaussian process metamodel,”
   Reliability Engineering & System Safety,
   vol. 94, no. 3, pp. 742–751, 2009.
   DOI:10.1016/j.ress.2008.07.008
"""

import numpy as np

from uqtestfuns.core.custom_typing import (
    MarginalSpecs,
    ProbInputSpecs,
    FunParamSpecs,
)
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["Ishigami"]


MARGINALS_ISHIGAMI1991: MarginalSpecs = [
    {
        "name": "X1",
        "distribution": "uniform",
        "parameters": [-np.pi, np.pi],
        "description": None,
    },
    {
        "name": "X2",
        "distribution": "uniform",
        "parameters": [-np.pi, np.pi],
        "description": None,
    },
    {
        "name": "X3",
        "distribution": "uniform",
        "parameters": [-np.pi, np.pi],
        "description": None,
    },
]

AVAILABLE_INPUTS: ProbInputSpecs = {
    "Ishigami1991": {
        "function_id": "Ishigami",
        "description": (
            "Probabilistic input model for the Ishigami function "
            "from Ishigami and Homma (1991)."
        ),
        "marginals": MARGINALS_ISHIGAMI1991,
        "copulas": None,
    },
}

AVAILABLE_PARAMETERS: FunParamSpecs = {
    "Ishigami1991": {
        "function_id": "Ishigami",
        "description": (
            "Parameter set for the Ishigami function from Ishigami and Homma "
            "(1991)"
        ),
        "declared_parameters": [
            {
                "keyword": "a",
                "value": 7.0,
                "type": float,
                "description": None,
            },
            {
                "keyword": "b",
                "value": 0.1,
                "type": float,
                "description": None,
            },
        ],
    },
    "Sobol1999": {
        "function_id": "Ishigami",
        "description": (
            "Parameter set for the Ishigami function from Sobol' and Levitan "
            "(1999)"
        ),
        "declared_parameters": [
            {
                "keyword": "a",
                "value": 7.0,
                "type": float,
                "description": None,
            },
            {
                "keyword": "b",
                "value": 0.05,
                "type": float,
                "description": None,
            },
        ],
    },
}

DEFAULT_PARAMETERS_SELECTION = "Ishigami1991"


def evaluate(xx: np.ndarray, a: float, b: float) -> np.ndarray:
    """Evaluate the Ishigami function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        3-Dimensional input values given by N-by-3 arrays where
        N is the number of input values.
    a : float
        The first parameter of the Ishigami function.
    b : float
        The second parameter of the Ishigami function.

    Returns
    -------
    np.ndarray
        The output of the Ishigami function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    # Compute the Ishigami function
    term_1 = np.sin(xx[:, 0])
    term_2 = a * np.sin(xx[:, 1]) ** 2
    term_3 = b * xx[:, 2] ** 4 * np.sin(xx[:, 0])

    yy = term_1 + term_2 + term_3

    return yy


class Ishigami(UQTestFunFixDimABC):
    """An implementation of the Ishigami test function."""

    _tags = ["sensitivity"]
    _description = "Ishigami function from Ishigami and Homma (1991)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = AVAILABLE_PARAMETERS
    _default_parameters_id = DEFAULT_PARAMETERS_SELECTION

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with an implementation of the circular bar RS problem.

The two-dimensional function is a variant of the classic RS reliability problem
with one quadratic term [1].

References
----------
1. A. K. Verma, S. Ajit, and D. R. Karanki, “Structural Reliability,”
   in Reliability and Safety Engineering, in Springer Series
   in Reliability Engineering. London: Springer London, 2016, pp. 257–292.
   DOI: 10.1007/978-1-4471-6269-8_8
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs, FunParamSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC


__all__ = ["RSCircularBar"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Verma2016": {
        "function_id": "RSCircularBar",
        "description": (
            "Input model for the circular bar RS from Verma et al. (2016)"
        ),
        "marginals": [
            {
                "name": "Y",
                "distribution": "normal",
                "parameters": [250.0, 25.0],
                "description": "Material mean yield strength [MPa]",
            },
            {
                "name": "F",
                "distribution": "normal",
                "parameters": [70.0, 7.0],
                "description": "Force mean value [kN]",
            },
        ],
        "copulas": None,
    },
}

AVAILABLE_PARAMETERS: FunParamSpecs = {
    "Verma2016": {
        "function_id": "RSCircularBar",
        "description": (
            "Parameter set for the RS circular bar reliability problem "
            "from Verma et al. (2016)"
        ),
        "declared_parameters": [
            {
                "keyword": "bar_diameter",
                "value": 25.0,
                "type": float,
                "description": "Diameter of the circular bar [mm]",
            },
        ],
    },
}


def evaluate(xx: np.ndarray, bar_diameter: float) -> np.ndarray:
    """Evaluate the circular bar RS problem on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        A two-dimensional input values given by an N-by-2 array
        where N is the number of input values.
    bar_diameter : float
        The diameter of the bar in [mm].

    Returns
    -------
    np.ndarray
        The performance function of the problem.
        If negative, the system is in failed state.
        The output is a one-dimensional array of length N.
    """
    # Convert to SI units
    yy_ = xx[:, 0] * 1e6  # From [MPa] to [Pa]
    ff = xx[:, 1] * 1e3  # From [kN] to [N]
    diam = bar_diameter * 1e-3  # From [mm] to [m]

    # Compute the performance function
    yy = yy_ - ff / (np.pi * diam**2 / 4)

    return yy


class RSCircularBar(UQTestFunFixDimABC):
    """Concrete implementation of the circular bar RS reliability problem."""

    _tags = ["reliability"]
    _description = "RS problem as a circular bar from Verma et al. (2016)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = AVAILABLE_PARAMETERS
    _default_input_dimension = 2

    evaluate = staticmethod(evaluate)  # type: ignore



"""
This module contains a collection of the default basis functions
for the meta.
"""

import numpy as np


__all__ = ["BASIS_BY_ID"]


def linear(xx: np.ndarray) -> np.ndarray:
    """Compute the linear function on a set of 1-dimensional points."""
    return xx


def quadratic(xx: np.ndarray) -> np.ndarray:
    """Compute the quadratic function on a set of 1-dimensional points."""
    return xx**2


def cubic(xx: np.ndarray) -> np.ndarray:
    """Compute the cubic function on a set of 1-dimensional points."""
    return xx**3


def exponential(xx: np.ndarray) -> np.ndarray:
    """Compute the exponential function on a set of 1-dimensional points."""
    yy = (np.exp(xx) - 1) / (np.exp(1) - 1)

    return yy


def periodic(xx: np.ndarray) -> np.ndarray:
    """Compute the sine periodic function on a set of 1-dimensional points."""
    yy = 0.5 * np.sin(2 * np.pi * xx)

    return yy


def discontinuous(xx: np.ndarray) -> np.ndarray:
    """Compute the discontinuous function on a set of 1-dimensional points."""
    yy = np.zeros(xx.shape)
    yy[xx > 0.5] = 1.0

    return yy


def null(xx: np.ndarray) -> np.ndarray:
    """Compute the null function on a set of 1-dimensional points."""
    yy = np.zeros(xx.shape)

    return yy


def non_monotonic(xx: np.ndarray) -> np.ndarray:
    """Compute the parabola function on a set of 1-dimensional points."""
    yy = 4 * (xx - 0.5) ** 2

    return yy


def inverse(xx: np.ndarray) -> np.ndarray:
    """Compute the inverse function on a set of 1-dimensional points."""
    yy = (10 - 1 / 1.1) ** (-1) * (xx + 0.1) ** (-1)

    return yy


def trigonometric(xx: np.ndarray) -> np.ndarray:
    """Compute the trigonometric function on a set of 1-dimensional points."""
    yy = np.cos(xx)

    return yy


BASIS_BY_ID = {
    0: linear,
    1: quadratic,
    2: cubic,
    3: exponential,
    4: periodic,
    5: discontinuous,
    6: null,
    7: non_monotonic,
    8: inverse,
    9: trigonometric,
}



"""
Module with an implementation of the Sobol-G test function.

The Sobol'-G function is an M-dimensional scalar-valued function.
It was introduced in [1] for testing numerical integration algorithms
(e.g., quasi-Monte-Carlo; see also for instance [2] and [3]).
The current form (and name) was from [4] and used in the context of global
sensitivity analysis. There, the function was generalized by introducing
a set of parameters that determines the importance of each input variable.
Later on, it becomes a popular testing function for global sensitivity analysis
methods; see, for instances, [5], [6], [7], and [9].

There are several sets of parameters used in the literature.

Notes
-----
- The parameters used in [5] and [6] correspond to
  the parameter choice 3 in [3].

References
----------

1. Paul Bratley, Bennet L. Fox, and Harald Niederreiter, "Implementation and
   tests of low-discrepancy sequences," ACM Transactions on Modeling and
   Computer Simulation, vol. 2, no. 3, pp. 195-213, 1992.
   DOI:10.1145/146382.146385
2. I. Radović, I. M. Sobol’, and R. F. Tichy, “Quasi-Monte Carlo Methods for
   Numerical Integration: Comparison of Different Low Discrepancy Sequences,”
   Monte Carlo Methods and Applications, vol. 2, no. 1, pp. 1–14, 1996.
   DOI: 10.1515/mcma.1996.2.1.1.
3. I. M. Sobol’, “On quasi-Monte Carlo integrations,” Mathematics and Computers
   in Simulation, vol. 47, no. 2–5, pp. 103–112, 1998.
   DOI: 10.1016/S0378-4754(98)00096-2
4. A. Saltelli and I. M. Sobol’, “About the use of rank transformation in
   sensitivity analysis of model output,” Reliability Engineering
   & System Safety, vol. 50, no. 3, pp. 225–239, 1995.
   DOI: 10.1016/0951-8320(95)00099-2.
5. A. Marrel, B. Iooss, F. Van Dorpe, and E. Volkova, “An efficient
   methodology for modeling complex computer codes with Gaussian processes,”
   Computational Statistics & Data Analysis, vol. 52, no. 10,
   pp. 4731–4744, 2008.
   DOI: 10.1016/j.csda.2008.03.026
6. A. Marrel, B. Iooss, B. Laurent, and O. Roustant, “Calculations of Sobol
   indices for the Gaussian process metamodel,” Reliability Engineering &
   System Safety, vol. 94, no. 3, pp. 742–751, 2009.
    DOI: 10.1016/j.ress.2008.07.008
7. S. Kucherenko, B. Feil, N. Shah, and W. Mauntz, “The identification of
   model effective dimensions using global sensitivity analysis,”
   Reliability Engineering & System Safety, vol. 96, no. 4, pp. 440–449, 2011.
   DOI: 10.1016/j.ress.2010.11.003
8. T. Crestaux, J.-M. Martinez, O. Le Maître, and O. Lafitte, “Polynomial
   chaos expansion for uncertainties quantification and sensitivity analysis,”
   presented at the Fifth International Conference on Sensitivity Analysis
   of Model Output, 2007.
   Accessed: Jan. 25, 2023
   URL: http://samo2007.chem.elte.hu/lectures/Crestaux.pdf
9. X. Sun, B. Croke, A. Jakeman, S. Roberts, "Benchmarking Active Subspace
   methods of global sensitivity analysis against variance-based Sobol’
   and Morris methods with established test functions," Environmental Modelling
   & Software, vol. 149, p. 105310, 2022.
   DOI: 10.1016/j.envsoft.2022.105310
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs, FunParamSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunVarDimABC

__all__ = ["SobolG"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Saltelli1995": {
        "function_id": "SobolG",
        "description": (
            "Probabilistic input model for the Sobol'-G function "
            "from Saltelli and Sobol' (1995)"
        ),
        "marginals": [
            {
                "name": "X",
                "distribution": "uniform",
                "parameters": [0.0, 1.0],
                "description": None,
            },
        ],
        "copulas": None,
    },
}

DEFAULT_INPUT_SELECTION = "Saltelli1995"


def _get_params_saltelli_1995_1(input_dimension: int) -> np.ndarray:
    """Construct a parameter array for Sobol'-G according to example 1 in [4].

    Notes
    -----
    - The function was most probably first appear in its original form
      in [1] (without parameters).
    - With the selected parameters, all input variables are equally important.
    """
    yy = np.zeros(input_dimension)

    return yy


def _get_params_saltelli_1995_2(input_dimension: int) -> np.ndarray:
    """Construct a parameter array for Sobol'-G according to example 2 in [4].

    Notes
    -----
    - With the selected parameters, the first two input variables are
      important, one is moderately important, and the rest is non-influential.
    - Originally, the dimension is limited to 8-dimensions; if more dimensions
      are used then the remaining dimension is also non-influential.
    """
    yy = np.zeros(input_dimension)

    if input_dimension > 1:
        yy[1] = 0

    if input_dimension > 2:
        yy[2] = 3

    if input_dimension > 3:
        yy[3:] = 9

    return yy


def _get_params_saltelli_1995_3(input_dimension: int) -> np.ndarray:
    """Construct a parameter array for Sobol'-G according to example 3 in [4].

    Notes
    -----
    - With the selected parameters, the first input variable is the most
      important and the importance of the remaining variables is decreasing.
    - Originally, the dimension is limited to 20-dimensions; if more dimensions
      are used then the remaining dimension is also non-influential.
    - The parameter set is also used in [8].
    """
    yy = (np.arange(1, input_dimension + 1) - 1) / 2.0

    return yy


def _get_params_sobol_1998_1(input_dimension: int) -> np.ndarray:
    """Construct a parameter array for Sobol'-G according to choice 1 in [3].

    Notes
    -----
    - Using this choice of parameters, the supremum of the Sobol-G function
      grows exponentially as a function of dimension about 2^M.
    """
    yy = 0.01 * np.ones(input_dimension)

    return yy


def _get_params_sobol_1998_2(input_dimension: int) -> np.ndarray:
    """Construct a parameter array for Sobol'-G according to choice 2 in [3].

    Notes
    -----
    - Using this choice of parameters, the supremum of the Sobol-G function
      grows exponentially as a function of dimension about (1.5)^M;
      it's a bit slower than choice 1.
    """
    yy = np.ones(input_dimension)

    return yy


def _get_params_sobol_1998_3(input_dimension: int) -> np.ndarray:
    """Construct a parameter array for Sobol'-G according to choice 3 in [3].

    Notes
    -----
    - Using this choice of parameters, the supremum of the Sobol-G function
      grows linearly as a function of dimension, i.e., 1 + (M/2).
    """
    yy = np.arange(1, input_dimension + 1)

    return yy


def _get_params_sobol_1998_4(input_dimension: int) -> np.ndarray:
    """Construct a parameter array for Sobol-G according to choice 4 in [3].

    Notes
    -----
    - Using this choice of parameters, the supremum of the Sobol-G function
      is bounded at 1.0.
    """
    yy = np.arange(1, input_dimension + 1) ** 2

    return yy


def _get_params_kucherenko_2011_2a(input_dimension: int) -> np.ndarray:
    """Construct a param. array for Sobol'-G according to problem 2A in [7]."""
    yy = np.zeros(input_dimension)
    if input_dimension >= 2:
        yy[2:] = 6.52

    return yy


def _get_params_kucherenko_2011_3b(input_dimension: int) -> np.ndarray:
    """Construct a parameter array for Sobol'-G according to problem 3B in [7].

    Notes
    -----
    - Using this choice of parameters, the supremum of the Sobol-G function
      grows exponentially as a function of dimension about (1.13)^M.
    """
    yy = 6.52 * np.ones(input_dimension)

    return yy


def _get_params_sun_2022(input_dimension: int) -> np.ndarray:
    """Construct a parameter array for Sobol'-G according to Sec. 3. 2. [9].

    Notes
    -----
    - The length of the array in [9] is originally seven following
      a 7-dimensional function. If input dimension is larger than 7,
      then the remaining coefficients are extrapolated
      from the last available value.
    """
    aa = np.array([0.0, 1.0, 4.5, 9.0, 99, 99, 99])

    if input_dimension > len(aa):
        aa_ext = 99 * np.ones(input_dimension - len(aa))
        aa = np.append(aa, aa_ext)

    return aa[:input_dimension]


AVAILABLE_PARAMETERS: FunParamSpecs = {
    "Saltelli1995-1": {
        "function_id": "SobolG",
        "description": (
            "Parameter set for the Sobol-G function from Saltelli and Sobol' "
            "(1995), example 1; all input variables are equally important"
        ),
        "declared_parameters": [
            {
                "keyword": "aa",
                "value": _get_params_saltelli_1995_1,
                "type": np.ndarray,
                "description": "Coefficients 'a'",
            },
        ],
    },
    "Saltelli1995-2": {
        "function_id": "SobolG",
        "description": (
            "Parameter set for the Sobol-G function from Saltelli and Sobol' "
            "(1995), example 2; the first two input variables are the most "
            "important and the rest are non-influential"
        ),
        "declared_parameters": [
            {
                "keyword": "aa",
                "value": _get_params_saltelli_1995_2,
                "type": np.ndarray,
                "description": "Coefficients 'a'",
            },
        ],
    },
    "Saltelli1995-3": {
        "function_id": "SobolG",
        "description": (
            "Parameter set for the Sobol-G function from Saltelli and Sobol' "
            "(1995), example 3; the first input variable is the most "
            "important and the importance of the remaining variables is "
            "decreasing"
        ),
        "declared_parameters": [
            {
                "keyword": "aa",
                "value": _get_params_saltelli_1995_3,
                "type": np.ndarray,
                "description": "Coefficients 'a'",
            },
        ],
    },
    "Sobol1998-1": {
        "function_id": "SobolG",
        "description": (
            "Parameter set for the Sobol-G function from Sobol' (1998)"
            "(1995), choice 1; the supremum of the function grows "
            "exponentially as a function of dimension about 2^M"
        ),
        "declared_parameters": [
            {
                "keyword": "aa",
                "value": _get_params_sobol_1998_1,
                "type": np.ndarray,
                "description": "Coefficients 'a'",
            },
        ],
    },
    "Sobol1998-2": {
        "function_id": "SobolG",
        "description": (
            "Parameter set for the Sobol-G function from Sobol' (1998)"
            "(1995), choice 2; the supremum of the function grows "
            "exponentially as a function of dimension about (1.5)^M"
        ),
        "declared_parameters": [
            {
                "keyword": "aa",
                "value": _get_params_sobol_1998_2,
                "type": np.ndarray,
                "description": "Coefficients 'a'",
            },
        ],
    },
    "Sobol1998-3": {
        "function_id": "SobolG",
        "description": (
            "Parameter set for the Sobol-G function from Sobol' (1998)"
            "(1995), choice 3; the supremum of the function grows "
            "linearly as a function of dimension, i.e., 1 + M / 2"
        ),
        "declared_parameters": [
            {
                "keyword": "aa",
                "value": _get_params_sobol_1998_3,
                "type": np.ndarray,
                "description": "Coefficients 'a'",
            },
        ],
    },
    "Sobol1998-4": {
        "function_id": "SobolG",
        "description": (
            "Parameter set for the Sobol-G function from Sobol' (1998)"
            "(1995), choice 4; the supremum of the function is bounded at 1.0"
        ),
        "declared_parameters": [
            {
                "keyword": "aa",
                "value": _get_params_sobol_1998_4,
                "type": np.ndarray,
                "description": "Coefficients 'a'",
            },
        ],
    },
    "Kucherenko2011-2a": {
        "function_id": "SobolG",
        "description": (
            "Parameter set for the Sobol-G function from Kucherenko et al. "
            "(2011), problem 2A"
        ),
        "declared_parameters": [
            {
                "keyword": "aa",
                "value": _get_params_kucherenko_2011_2a,
                "type": np.ndarray,
                "description": "Coefficients 'a'",
            },
        ],
    },
    "Kucherenko2011-3b": {
        "function_id": "SobolG",
        "description": (
            "Parameter set for the Sobol-G function from Kucherenko et al. "
            "(2011), problem 3B; the supremum of the function grows "
            "exponentially as a function of dimension about (1.13)^M"
        ),
        "declared_parameters": [
            {
                "keyword": "aa",
                "value": _get_params_kucherenko_2011_3b,
                "type": np.ndarray,
                "description": "Coefficients 'a'",
            },
        ],
    },
    "Sun2022": {
        "function_id": "SobolG",
        "description": (
            "Parameter set for the Sobol-G function from Sun et al. "
            "(2022), Section 3.2"
        ),
        "declared_parameters": [
            {
                "keyword": "aa",
                "value": _get_params_sun_2022,
                "type": np.ndarray,
                "description": "Coefficients 'a'",
            },
        ],
    },
}

DEFAULT_PARAMETERS_SELECTION = "Saltelli1995-3"


def evaluate(xx: np.ndarray, aa: np.ndarray):
    """Evaluate the Sobol-G function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        M-Dimensional input values given by an N-by-M array where
        N is the number of input values.
    aa : np.ndarray
        The vector of parameters (i.e., coefficients) of the Sobol'-G function;
        the length of the vector is the same as the number of input dimensions.

    Returns
    -------
    np.ndarray
        The output of the Sobol-G function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = np.prod(((np.abs(4 * xx - 2) + aa) / (1 + aa)), axis=1)

    return yy


class SobolG(UQTestFunVarDimABC):
    """An implementation of the M-dimensional Sobol'-G test function."""

    _tags = ["sensitivity", "integration"]
    _description = "Sobol'-G function from Saltelli and Sobol' (1995)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = AVAILABLE_PARAMETERS
    _default_parameters_id = DEFAULT_PARAMETERS_SELECTION

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with routines involving the Gumbel maximum probability distribution.

The Gumbel distribution in UQTestFuns is parameterized by two parameters:
``mu`` and ``beta``.

The underlying implementation is based on the implementation from scipy.stats
(specifically, ``gumbel_r``  the right-skewed Gumbel).
In the SciPy convention, the parameter ``mu`` corresponds
to the ``loc`` parameter, while the parameter ``beta``
corresponds to the ``scale`` parameter.
"""

import numpy as np

from scipy.stats import gumbel_r

from .utils import verify_param_nums, postprocess_icdf
from ....global_settings import ARRAY_FLOAT

DISTRIBUTION_NAME = "gumbel"

NUM_PARAMS = 2


def verify_parameters(parameters: ARRAY_FLOAT) -> None:
    """Verify the parameters of a Gumbel (max.) distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a Gumbel (max.) distribution
        (i.e., the mu and beta parameters).

    Returns
    ------
    None
        The function exits without any return value when nothing is wrong.

    Raises
    ------
    ValueError
        If any of the parameter values are invalid
        or the shapes are inconsistent.
    """
    # Verify overall shape
    verify_param_nums(parameters.size, NUM_PARAMS, DISTRIBUTION_NAME)

    if parameters[1] <= 0.0:
        raise ValueError(
            f"The corresponding beta parameter {parameters[1]} "
            f"of a {DISTRIBUTION_NAME} must be larger than 0.0!"
        )


def lower(parameters: ARRAY_FLOAT):
    """Get the lower bound of a Gumbel (max.) distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a Gumbel (max.) distribution.

    Returns
    -------
    float
        The lower bound of the Gumbel (max.) distribution.

    Notes
    -----
    - Strictly speaking, a Gumbel (max.) distribution is unbounded on both
      sides. However, for numerical reason a lower bound is set.
    - The probability mass contained up to the lower bound of the Gumbel (max.)
      distribution is at most 1e-15.
    - The difference between 1.0 and the probability mass between lower and
      upper bounds is at most 1e-15.
    """
    # 3.606621167487737 is the quantile values with probability of 1e-16
    # for the standard Gumbel (max.) distribution (mu = 0.0, beta = 1.0)
    lower_bound = float(parameters[0] - 3.606621167487737 * parameters[1])

    return lower_bound


def upper(parameters: ARRAY_FLOAT):
    """Get the upper bound of a Gumbel (max.) distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a Gumbel (max.) distribution.

    Returns
    -------
    float
        The upper bound of the Gumbel (max.) distribution.

    Notes
    -----
    - Strictly speaking, a Gumbel (max.) distribution is unbounded on both
      sides. However, for numerical reason a lower bound is set.
    - The difference between 1.0 and probability mass contained up to the
      upper bound of the Gumbel (max.) distribution is smaller than 1e-15.
    - The difference between 1.0 and the probability mass between lower and
      upper bounds is smaller than 1e-15.
    """
    # 36.7368005696771 is the quantile values with probability of 1-1e-16
    # for the standard Gumbel (max.) distribution (mu = 0.0, beta = 1.0)
    upper_bound = float(parameters[0] + 36.7368005696771 * parameters[1])

    return upper_bound


def pdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the PDF values of a Gumbel (max.) distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of a Gumbel (max.) distribution.
    parameters : ARRAY_FLOAT
        Parameters of the Gumbel (max.) distribution.
    lower_bound : float
        Lower bound of the Gumbel (max.) distribution.
    upper_bound : float
        Upper bound of the Gumbel (max.) distribution.

    Returns
    -------
    ARRAY_FLOAT
        PDF values of the Gumbel (max.) distribution on the sample values.

    Notes
    -----
    - The PDF for sample with values outside the bounds are set to 0.0.
    """
    idx_non_zero = np.logical_and(xx >= lower_bound, xx <= upper_bound)

    # Get the parameters
    mu = parameters[0]
    beta = parameters[1]

    yy = np.zeros(xx.shape)
    yy[idx_non_zero] = gumbel_r.pdf(xx[idx_non_zero], loc=mu, scale=beta)

    return yy


def cdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the CDF values of a Gumbel (max.) distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of a Gumbel (max.) distribution.
    parameters : ARRAY_FLOAT
        Parameters of the Gumbel (max.) distribution.
    lower_bound : float
        Lower bound of the Gumbel (max.) distribution.
    upper_bound : float
        Upper bound of the Gumbel (max.) distribution.

    Returns
    -------
    ARRAY_FLOAT
        CDF values of the Gumbel (max.) distribution on the sample values.

    Notes
    -----
    - CDF for sample with values smaller (resp. larger) than the lower bound
      (resp. upper bound) are set to 0.0 (resp. 1.0).
    """

    # Get the parameters
    mu = parameters[0]
    beta = parameters[1]

    yy = np.empty(xx.shape)

    # Set CDF to 0.0 for values below the lower bound
    idx_lower = xx < lower_bound
    yy[idx_lower] = 0.0

    # Set CDF to 1.0 for values above the upper bound
    idx_upper = xx > upper_bound
    yy[idx_upper] = 1.0

    idx_rest = np.logical_and(
        np.logical_not(idx_lower), np.logical_not(idx_upper)
    )
    yy[idx_rest] = gumbel_r.cdf(xx[idx_rest], loc=mu, scale=beta)

    return yy


def icdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the ICDF values of a Gumbel (max.) distribution.

    Parameters
    ----------
    xx : np.ndarray
        Sample values (realizations) in the [0, 1] domain.
    parameters : np.ndarray
        Parameters of a Gumbel (max.) distribution.
    lower_bound : float
        Lower bound of a Gumbel (max.) distribution.
    upper_bound : float
        Upper bound of a Gumbel (max.) distribution.

    Returns
    -------
    ARRAY_FLOAT
        Transformed values in the domain of the Gumbel (max.) distribution.

    Notes
    -----
    - ICDF for sample values outside [0.0, 1.0] is set to NaN according to
      the underlying SciPy implementation.
    """
    # Get the parameters
    mu = parameters[0]
    beta = parameters[1]

    # Compute the ICDF
    yy = gumbel_r.ppf(xx, loc=mu, scale=beta)

    # Check if values are within the set bounds
    yy = postprocess_icdf(yy, lower_bound, upper_bound)

    return yy



"""
Module with an implementation of the Forrester et al. (2008) test function.

The Forrester2008 test function is a one-dimensional scalar-valued function.
The function is multimodal, with a single global minimum, a single local
minimum, and a zero-gradient at the inflection point.

References
----------

1. William J. Welch, Robert J. Buck, Jerome Sacks, Henry P. Wynn,
   Toby J. Mitchell, and Max D. Morris, "Screening, predicting, and computer
   experiments," Technometrics, vol. 34, no. 1, pp. 15-25, 1992.
   DOI: 10.2307/1269548
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["Forrester2008"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Forrester2008": {
        "function_id": "Forrester2008",
        "description": (
            "Input specification for the 1D test function "
            "from Forrester et al. (2008)"
        ),
        "marginals": [
            {
                "name": "x",
                "distribution": "uniform",
                "parameters": [0, 1],
                "description": None,
            },
        ],
        "copulas": None,
    },
}


def evaluate(xx: np.ndarray) -> np.ndarray:
    """The evaluation function for the Forrester et al. (2008) function.

    Parameters
    ----------
    xx : np.ndarray
        One-dimensional input values given by an N-by-1 array
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the 1D Forrester et al. (2008) function evaluated
        on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = (6 * xx[:, 0] - 2) ** 2 * np.sin(12 * xx[:, 0] - 4)

    return yy


class Forrester2008(UQTestFunFixDimABC):
    """An implementation of the 1D function of Forrester et al. (2008)."""

    _tags = ["optimization", "metamodeling"]
    _description = "One-dimensional function from Forrester et al. (2008)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None

    evaluate = staticmethod(evaluate)  # type: ignore



"""
This module contains the classes that implement specification
for a meta and the resulting test function.
"""

import numpy as np
import itertools
from dataclasses import dataclass, field, InitVar
from scipy.special import comb
from typing import Dict, Callable, Tuple, Optional, Union, List

from ..core import Marginal

__all__ = ["UQMetaFunSpec", "UQTestFunSpec"]


def _preprocess_effects(
    effects_dict: Dict[int, Optional[int]], input_dimension: int
) -> Dict[int, int]:
    """Preprocess the effects dictionary.

    It will do the following things:
    - If a value is None, replace it with the number of all possible
      interaction terms.
    - If a value is larger than the dimension of the test function
      or it's zero, then exclude the key.

    >>> effects_dict = {1: None, 2: 3, 3: None, 4: 10}
    >>> input_dimension = 3
    >>> _preprocess_effects(effects_dict, input_dimension)
    {1: 3, 2: 3, 3: 1}

    Parameters
    ----------
    effects_dict : Dict[int, Optional[int]]
        Specified effects with the corresponding length to take into account
        in a realization of test function.
    input_dimension : int
        Number of dimensions of the test function.

    Returns
    -------
    Dict[int, int]
        Cleaned-up specified effects.
    """
    effects = dict()

    for key in list(effects_dict):
        if key <= input_dimension and effects_dict[key] != 0:
            if effects_dict[key] is None:
                # NOTE: scipy.special.comb returns a flot
                effects[key] = int(comb(input_dimension, key))
            else:
                effects[key] = effects_dict[key]  # type: ignore

    return effects


def _select_basis(input_dimension: int, num_basis: int) -> Tuple[int, ...]:
    """Select basis of a certain length from a selection of basis.

    For example, with input dimension of 3 and number of basis functions
    of 9, the output may be a tuple of integers: (7, 3, 8) which means
    the first, second, and third dimension has the 7th, 3rd,
    and 8th basis function, respectively.

    Parameters
    ----------
    input_dimension : int
        Number of dimensions of the test function.
    num_basis : int
        Number of available basis functions to select from.

    Returns
    -------
    Tuple[int, ...]
        Selected basis function in each dimension.
    """
    selection = tuple(np.random.randint(0, num_basis, size=input_dimension))

    return selection


def _create_effects_tuples(
    input_dimension: int, effects_dict: Dict[int, int]
) -> Dict[int, Tuple[Tuple[int, ...], ...]]:
    """Create a dictionary of effect matrices.

    >>> input_dimension = 3
    >>> effects_dict = {1: None, 2: None, 3: None}
    >>> _create_effects_tuples(input_dimension, effects_dict)
    {1: ((0,), (1,), (2,)), 2: ((0, 1), (0, 2), (1, 2)), 3: ((0, 1, 2),)}
    >>> effects_dict = {}
    >>> _create_effects_tuples(input_dimension, effects_dict)
    {}
    >>> effects_dict = {1: 0, 2: None, 3: 0}
    >>> _create_effects_tuples(input_dimension, effects_dict)
    {2: ((0, 1), (0, 2), (1, 2))}
    >>> effects_dict = {5: None}
    >>> _create_effects_tuples(input_dimension, effects_dict)
    {}

    Parameters
    ----------
    input_dimension : int
        Number of dimensions of the test function.
    effects_dict : Dict[int, int]
        Specified effects to take into account.

    Returns
    -------
    Dict[int, Tuple[Tuple[int, ...], ...]]
        Effects tuples for each n-way interaction as specified in the effects.

    Notes
    -----
    - If an effect has a value of 0 then no effect is taken into account.
    - If an effect is larger than the input dimension of the test function
      then no effect is taken into account.
    """
    effects_tuples = dict()

    for key in effects_dict:
        if (
            effects_dict[key] != 0 and key <= input_dimension
        ):  # pragma: no cover
            effects_tuples[key] = tuple(
                itertools.combinations(np.arange(input_dimension), key)
            )

    return effects_tuples


def _select_effects(
    all_effects: Dict[int, Tuple[Tuple[int, ...], ...]],
    effects_dict: Dict[int, int],
) -> Dict[int, Tuple[Tuple[int, ...], ...]]:
    """Randomly select effects from all possible terms.

    Parameters
    ----------
    all_effects : Dict[int, Tuple[Tuple[int, ...], ...]]
        Effects tuples for each n-way interaction as specified in the effects.
    effects_dict : Dict[int, int]
        Specified effects with the corresponding length to take into account in
        a realization of test function.

    Returns
    -------
    Dict[int, Tuple[Tuple[int, ...], ...]]
        Selected interaction terms for each n-way interactions.
    """
    selected_effects = dict()

    for key in effects_dict:
        length = effects_dict[key]

        if length == len(all_effects[key]):
            # Take all
            selected_effects[key] = all_effects[key]
        else:
            # Randomly select the interaction tuples
            idx = np.random.choice(
                len(all_effects[key]), length, replace=False
            )
            selected_effects[key] = tuple([all_effects[key][i] for i in idx])

    return selected_effects


def _generate_effect_coeffs(
    coeffs_gen: Callable, effects_dict: Dict[int, int]
) -> Dict[int, np.ndarray]:
    """Generate the coefficient values for each effect term.

    Parameters
    ----------
    coeffs_gen : Callable
        Function to generate the coefficient values for each effect term.
    effects_dict : Dict[int, int]
        Specified effects with the corresponding length to take into account in
        a realization of test function.

    Returns
    -------
    Dict[int, np.ndarray]
        The coefficient values of each effect term.
    """
    coeffs = dict()

    for key in effects_dict:
        coeffs[key] = coeffs_gen(effects_dict[key])

    return coeffs


def _select_marginals(
    marginals: Union[List[Dict], Tuple[Dict, ...]], input_dimension: int
) -> Union[List[Dict], Tuple[Dict, ...]]:
    """Randomly select marginals of a given dimension from a set of inputs.
    # TODO: Update the description and the type hints
    Parameters
    ----------
    inputs : Union[List[Dict], Tuple[Dict, ...]]
        List of available marginals to construct a probabilistic input model
        for a test function realization.
    input_dimension : int
        Number of dimensions of the test function.

    Returns
    -------
    Union[List[Dict], Tuple[Dict, ...]]
        List of selected marginals to construct a probabilistic input model of
        the given dimension for a test function realization.
    """
    selected_marginals = []
    indices = np.random.randint(
        low=0, high=len(marginals), size=input_dimension
    )

    for num, idx in enumerate(indices):
        selected_marginal = marginals[idx]
        name = getattr(selected_marginal, "name", None)
        if name is None:
            name = f"X{num + 1}"
        # Create a new instance (now with a name)
        selected_marginal = Marginal(
            name=name,
            distribution=selected_marginal.distribution,
            parameters=selected_marginal.parameters,
            description=selected_marginal.description,
        )

        selected_marginals.append(selected_marginal)

    return selected_marginals


@dataclass(frozen=True)
class UQTestFunSpec:
    """Specification class for a test function realization.

    An instance of UQTestFunSpec can be used to create a realization of test
    function from a fully-specified meta.

    Parameters
    ----------
    input_dimension : int
        Number of dimension of the test function.
    basis_functions : Dict[int, Callable]
        Collection of basis functions to choose from.
    selected_basis : Tuple[int, ...]
        Selected basis functions to construct the test function.
    effects_tuples : Dict[int, Tuple[Tuple[int, ...], ...]]
        Tuples that codify cross-terms between basis functions
        of each dimension that represents interaction terms
        in the test function.
    effects_coeffs : Dict[int, np.ndarray]
        Coefficients that appear in each of the terms in the test function.
    inputs : Union[List[Dict], Tuple[Dict, ...]]  # TODO: Update this
        Input marginals to construct a multi-dimensional probabilistic input of
        the test function.
    """

    input_dimension: int
    basis_functions: Dict[int, Callable]
    selected_basis: Tuple[int, ...]
    effects_tuples: Dict[int, Tuple[Tuple[int, ...], ...]]
    effects_coeffs: Dict[int, np.ndarray]
    inputs: Union[List[Marginal], Tuple[Marginal, ...]]


@dataclass
class UQMetaFunSpec:
    """Specification class for a meta.

    Parameters
    ----------
    input_dimension : int
        Number of dimension of the generated test functions.
    basis_functions : Dict[int, Callable]
        Collection of basis functions to choose from.
    effects : Dict[int, Optional[int]]
        Specified effects with the corresponding length to take into account
        for all the test function realizations.
    inputs : Union[List[Dict], Tuple[Dict, ...]]  # TODO: Update this
        List of available input marginals to construct a multi-dimensional
        probabilistic input of the test function realizations.
    coeffs_generator : Callable
        Function to generate the coefficient values for each effect term.
    """

    input_dimension: int
    basis_functions: Dict[int, Callable]
    effects: Dict[int, int] = field(init=False)
    effects_dict: InitVar[Dict[int, Optional[int]]]
    input_marginals: Union[List[Marginal], Tuple[Marginal, ...]]
    coeffs_generator: Callable
    _effects_tuples: Optional[Dict[int, Tuple[Tuple[int, ...], ...]]] = field(
        init=False, repr=False
    )

    def __post_init__(self, effects_dict):
        if self.input_dimension < 1:
            raise ValueError(
                f"input dimension must be > 0! Got {self.input_dimension}"
            )
        self._effects_tuples = None
        # Clean up the effects dictionary
        self.effects = _preprocess_effects(effects_dict, self.input_dimension)

    def get_sample(
        self, sample_size: int = 1
    ) -> Optional[Union[UQTestFunSpec, List[UQTestFunSpec]]]:
        """Get realizations of UQTestFunSpec.

        Parameters
        ----------
        sample_size : int, Optional
            Number of realizations.

        Returns
        -------
        Optional[Union[UQTestFunSpec, List[UQTestFunSpec]]]
            Realization(s) of UQTestFunSpec.

        Notes
        -----
        - An instance of UQTestFunSpec is used to construct a realization
          of a test function from the given meta.
        - With sample_size larger than one, a list of UQTestFunSpec instances
          is returned.
        """

        if sample_size < 1:
            return None

        # Generate all possible requested interactions and cache them
        if self._effects_tuples is None:
            self._effects_tuples = _create_effects_tuples(
                self.input_dimension, self.effects
            )

        sample = []
        for _ in range(sample_size):
            # Randomly select basis
            num_basis = len(self.basis_functions)
            selected_basis = _select_basis(self.input_dimension, num_basis)

            # Randomly select tuples of interaction terms
            effects_tuples = _select_effects(
                self._effects_tuples, self.effects
            )

            # Randomly generate the coefficients for all terms
            effects_coeffs = _generate_effect_coeffs(
                self.coeffs_generator, self.effects
            )

            # Randomly select input marginals and create an instance
            selected_input_marginals = _select_marginals(
                self.input_marginals, self.input_dimension
            )

            # Create an instance of UQTestFunSpec
            uqtestfun_spec = UQTestFunSpec(
                self.input_dimension,
                self.basis_functions,
                selected_basis,
                effects_tuples,
                effects_coeffs,
                selected_input_marginals,
            )

            sample.append(uqtestfun_spec)

        if sample_size == 1:
            return sample[0]
        else:
            return sample


if __name__ == "__main__":  # pragma: no cover
    import doctest

    doctest.testmod()



"""
The test module for the high-level function 'list_functions()'
"""

import pytest

from conftest import assert_call

from uqtestfuns import list_functions, test_functions
from uqtestfuns.utils import get_available_classes, SUPPORTED_TAGS


def test_default_call():
    """Test function call without any arguments."""
    assert_call(list_functions)


class TestValidArgument:
    """A series of tests for calling list_functions() with valid arguments."""

    @pytest.mark.parametrize("input_dimension", [1, 2, 10, "M", None])
    def test_input_dimension(self, input_dimension):
        """Test function call with 'input_dimension' argument."""
        assert_call(list_functions, input_dimension=input_dimension)

    @pytest.mark.parametrize("tag", SUPPORTED_TAGS)
    def test_tag(self, tag):
        """Test function call with 'tag' argument."""
        assert_call(list_functions, tag=tag)

    @pytest.mark.parametrize("output_dimension", [1, 2, 3, None])
    def test_output_dimension(self, output_dimension):
        """Test function call with 'tag' argument."""
        assert_call(list_functions, output_dimension=output_dimension)

    @pytest.mark.parametrize("parameterized", [True, False, None])
    def test_parameterized(self, parameterized):
        """Test function call with 'tag' argument."""
        assert_call(list_functions, parameterized=parameterized)

    @pytest.mark.parametrize("tabulate", [True, False, None])
    def test_tabulate(self, tabulate):
        """Test function call with 'tag' argument."""
        assert_call(list_functions, tabulate=tabulate)


class TestInvalidValueArgument:
    """A series of tests for calling list_functions() w/ invalid value arg."""

    @pytest.mark.parametrize("input_dimension", [-1, -2, -3, "a"])
    def test_input_dimension(self, input_dimension):
        """Test function call with invalid value for 'input_dimension'."""
        with pytest.raises(ValueError):
            list_functions(input_dimension=input_dimension)

    @pytest.mark.parametrize("tag", ["hello", "world"])
    def test_tag(self, tag):
        """Test function call with invalid value for 'tag'."""
        with pytest.raises(ValueError):
            list_functions(tag=tag)

    @pytest.mark.parametrize("output_dimension", [-1, -2, -3])
    def test_output_dimension(self, output_dimension):
        """Test function call with invalid value for 'output_dimension'."""
        with pytest.raises(ValueError):
            list_functions(output_dimension=output_dimension)


class TestInvalidTypeArgument:
    """A series of tests for calling list_functions() w/ invalid type arg."""

    @pytest.mark.parametrize("input_dimension", [1.0, [2]])
    def test_input_dimension(self, input_dimension):
        """Test function call with invalid type for 'input_dimension'."""
        with pytest.raises(TypeError):
            list_functions(input_dimension=input_dimension)

    @pytest.mark.parametrize("tag", [1.0, 5.0, False])
    def test_tag(self, tag):
        """Test function call with invalid type for 'tag'."""
        with pytest.raises(TypeError):
            list_functions(tag=tag)

    @pytest.mark.parametrize("output_dimension", ["a", [2]])
    def test_output_dimension(self, output_dimension):
        """Test function call with invalid type for 'output_dimension'."""
        with pytest.raises(TypeError):
            list_functions(output_dimension=output_dimension)

    @pytest.mark.parametrize("parameterized", ["a", 1, 2])
    def test_parameterized(self, parameterized):
        """Test function call with invalid type for 'parameterized'."""
        with pytest.raises(TypeError):
            list_functions(parameterized=parameterized)

    @pytest.mark.parametrize("tabulate", [1.0, 100, "100"])
    def test_tabulate(self, tabulate):
        """Test function call with invalid type for 'tabulate'."""
        with pytest.raises(TypeError):
            list_functions(tabulate=tabulate)


def test_untabulated_call():
    """Test function call that returns list of classes."""

    # --- Arrange
    tabulate = False

    # --- Act
    my_classes_from_list = list_functions(tabulate=tabulate)
    my_classes_ref = dict(get_available_classes(test_functions))

    # --- Assertions
    assert isinstance(my_classes_from_list, list)
    assert len(my_classes_from_list) == len(my_classes_ref)
    for my_class in my_classes_from_list:
        assert my_class in list(my_classes_ref.values())


def test_tablefmt_html():
    """Test function call with 'html' as tablefmt."""

    table = list_functions(tablefmt="html")

    # Assertion
    assert isinstance(table, str)



"""
Module with an implementation of the Piston simulation test function.

The Piston simulation test function is a seven-dimensional scalar-valued
test function.
The function computes the cycle time of a piston.
The function has been used as a test function in metamodeling exercises [1].
A 20-dimensional variant was used for sensitivity analysis in [2]
by introducing 13 additional inert input variables.

References
----------

1. E. N. Ben-Ari and D. M. Steinberg, "Modeling data from computer
   experiments: An empirical comparison of Kriging with MARS
   and projection pursuit regression," Quality Engineering,
   vol. 19, pp. 327-338, 2007.
   DOI: 10.1080/08982110701580930
2. H. Moon, "Design and Analysis of Computer Experiments for Screening Input
   Variables," Ph. D. dissertation, Ohio State University, Ohio, 2010.
   URL: http://rave.ohiolink.edu/etdc/view?acc_num=osu1275422248
"""

import numpy as np

from copy import deepcopy

from uqtestfuns.core.custom_typing import MarginalSpecs, ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["Piston"]


# Marginals specification from [1]
MARGINALS_BENARI2007: MarginalSpecs = [
    {
        "name": "M",
        "distribution": "uniform",
        "parameters": [30.0, 60.0],
        "description": "Piston weight [kg]",
    },
    {
        "name": "S",
        "distribution": "uniform",
        "parameters": [0.005, 0.020],
        "description": "Piston surface area [m^2]",
    },
    {
        "name": "V0",
        "distribution": "uniform",
        "parameters": [0.002, 0.010],
        "description": "Initial gas volume [m^3]",
    },
    {
        "name": "k",
        "distribution": "uniform",
        "parameters": [1000.0, 5000.0],
        "description": "Spring coefficient [N/m]",
    },
    {
        "name": "P0",
        "distribution": "uniform",
        "parameters": [90000.0, 110000.0],
        "description": "Atmospheric pressure [N/m^2]",
    },
    {
        "name": "Ta",
        "distribution": "uniform",
        "parameters": [290.0, 296.0],
        "description": "Ambient temperature [K]",
    },
    {
        "name": "T0",
        "distribution": "uniform",
        "parameters": [340.0, 360.0],
        "description": "Filling gas temperature [K]",
    },
]

# Marginals specification from [2]
MARGINALS_MOON2010 = [deepcopy(_) for _ in MARGINALS_BENARI2007]
for i in range(13):
    MARGINALS_MOON2010.append(
        {
            "name": f"Inert {i+1}",
            "distribution": "uniform",
            "parameters": [100.0, 200.0],
            "description": "Inert input [-]",
        }
    )

AVAILABLE_INPUTS: ProbInputSpecs = {
    "BenAri2007": {
        "function_id": "Piston",
        "description": (
            "Probabilistic input model for the Piston simulation model "
            "from Ben-Ari and Steinberg (2007)."
        ),
        "marginals": MARGINALS_BENARI2007,
        "copulas": None,
    },
    "Moon2010": {
        "function_id": "Piston",
        "description": (
            "Probabilistic input model for the Piston simulation model "
            "from Moon (2010)."
        ),
        "marginals": MARGINALS_MOON2010,
        "copulas": None,
    },
}

DEFAULT_INPUT_SELECTION = "BenAri2007"


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the Piston simulation test function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        (At least) 6-dimensional input values given by N-by-6 arrays
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the Piston simulation test function,
        The output is a one-dimensional array of length N.

    Notes
    -----
    - The variant of this test function has 13 additional inputs,
      but they are all taken to be inert and therefore should not affect
      the output.
    """
    mm = xx[:, 0]  # piston weight
    ss = xx[:, 1]  # piston surface area
    vv_0 = xx[:, 2]  # initial gas volume
    kk = xx[:, 3]  # spring coefficient
    pp_0 = xx[:, 4]  # atmospheric pressure
    tt_a = xx[:, 5]  # ambient temperature
    tt_0 = xx[:, 6]  # filling gas temperature

    # Compute the force
    aa = pp_0 * ss + 19.62 * mm - kk * vv_0 / ss

    # Compute the force difference
    daa = np.sqrt(aa**2 + 4.0 * kk * pp_0 * vv_0 * tt_a / tt_0) - aa

    # Compute the volume difference
    vv = ss / 2.0 / kk * daa

    # Compute the cycle time
    cc = (
        2.0
        * np.pi
        * np.sqrt(mm / (kk + ss**2 * pp_0 * vv_0 * tt_a / tt_0 / vv**2))
    )

    return cc


class Piston(UQTestFunFixDimABC):
    """A concrete implementation of the Piston simulation test function."""

    _tags = ["metamodeling", "sensitivity"]
    _description = "Piston simulation model from Ben-Ari and Steinberg (2007)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None
    _default_input_id = DEFAULT_INPUT_SELECTION

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with routines involving the truncated normal probability distribution.

The truncated normal distribution in UQTestFuns is parameterized by four
parameters: ``mu``, ``sigma``, ``lb``, and ``ub`` that correspond to
the mean, the standard deviation, lower bound, and upper bound, respectively.

The underlying implementation is based on the implementation of scipy.stats.
In the SciPy implementation, the truncated normal is parameterized also by
four parameters but with different meaning: a, b, loc, and scale. loc and scale
correspond to the mu and sigma of the underlying normal distribution, while
a and b correspond to the lower and upper bound, respectively.
However, the lower and upper bounds correspond to the standard normal
distribution.

One might think that, for instance, with loc = 0.5 and scale = 0.1,
and a = 0 and b = 1, the distribution will be centered on 0.5
and cut at 0 and 1.
Yet, this is not what the scipy.stats.beta does.
Instead, it is a truncated standard normal at 0 and 1 then shift to the new
``loc`` of 0.5; the lower and upper bounds are actually, 0.5 and 1.5,
respectively.

The translation between parameterization of the distribution in UQTestFuns
and SciPy is as follows:

- ``a`` = ``(lb - mu) / sigma``
- ``b`` = ``(ub - mu) / sigma``
- ``loc`` = ``mu``
- ``scale`` = ``sigma``

where lb and ub are the desired lower and upper bounds of the truncated
normal distribution, respectively.
"""

import numpy as np

from scipy.stats import truncnorm
from typing import Tuple

from . import normal
from .utils import verify_param_nums, postprocess_icdf
from ....global_settings import ARRAY_FLOAT

DISTRIBUTION_NAME = "trunc-normal"

NUM_PARAMS = 4


def _get_parameters(
    parameters: ARRAY_FLOAT,
) -> Tuple[float, float, float, float]:
    """Get the parameters of a truncated normal dist. w/ intuitive names.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a truncated normal distribution.

    Returns
    -------
    Tuple[float, float, float, float]
        The mean, standard deviation, lower bound, and upper bound of the
        truncated normal distribution.
    """
    mu = float(parameters[0])
    sigma = float(parameters[1])
    lb = float(parameters[2])
    ub = float(parameters[3])

    return mu, sigma, lb, ub


def verify_parameters(parameters: ARRAY_FLOAT) -> None:
    """Verify the parameters of a truncated normal distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a truncated normal distribution
        (i.e., the mean, standard deviation, and the lower and upper bounds).

    Returns
    ------
    None
        The function exits without any return value if nothing is wrong.

    Raises
    ------
    ValueError
        If any of the parameter values are invalid
        or the shapes are inconsistent.
    """
    # Verify overall shape
    verify_param_nums(parameters.size, NUM_PARAMS, DISTRIBUTION_NAME)

    mu, sigma, lb, ub = _get_parameters(parameters)

    # Check validity of values
    if sigma <= 0.0:
        raise ValueError(
            f"The standard deviation {sigma} "
            f"must be larger than 0.0 (positive)!"
        )

    if lb >= ub:
        raise ValueError(
            f"The lower bound {lb} "
            f"cannot be equal or greater than the upper bound {ub}!"
        )


def lower(parameters: ARRAY_FLOAT) -> float:
    """Get the lower bound of a truncated normal distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a truncated normal distribution.

    Returns
    -------
    float
        The lower bound of the truncated normal distribution.
    """
    _, _, lower_bound, _ = _get_parameters(parameters)

    # Get the lower bound of the normal distribution
    lb_normal = normal.lower(parameters[:2])
    if lower_bound < lb_normal:
        # Use the corresponding normal lower bound instead
        return lb_normal

    return lower_bound


def upper(parameters: ARRAY_FLOAT) -> float:
    """Get the upper bound of a truncated normal distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a truncated normal distribution

    Returns
    -------
    float
        The upper bound of the truncated normal distribution.
    """
    _, _, _, upper_bound = _get_parameters(parameters)

    # Get the upper bound of the normal distribution
    ub_normal = normal.upper(parameters[:2])
    if upper_bound > ub_normal:
        # Use the corresponding normal upper bound instead
        return ub_normal

    return upper_bound


def pdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the PDF values of a truncated normal distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of a truncated normal distribution.
    parameters : ARRAY_FLOAT
        Parameters of the truncated normal distribution.
    lower_bound : float
        Lower bound of the truncated normal distribution.
    upper_bound : float
        Upper bound of the truncated normal distribution.

    Returns
    -------
    ARRAY_FLOAT
        PDF values of the truncated normal distribution on the sample values.

    Notes
    -----
    - The values outside the bounds are set to 0.0.
    - ``lower_bound`` and ``upper_bound`` in the function parameters may not
      be the same as the ones in ``parameters`` as the former are already
      processed setting the numerical bounds when the given bounds are outside
      them.
    """
    yy = np.zeros(xx.shape)
    idx = np.logical_and(xx >= lower_bound, xx <= upper_bound)

    # NOTE: Don't use the original bounds as "lower_bound" and "upper_bound"
    # may have been further truncated for numerical reason.
    mu, sigma, _, _ = _get_parameters(parameters)

    yy[idx] = truncnorm.pdf(
        xx[idx],
        a=(lower_bound - mu) / sigma,
        b=(upper_bound - mu) / sigma,
        loc=mu,
        scale=sigma,
    )

    return yy


def cdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the CDF values of the truncated normal distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of a truncated normal distribution.
    parameters : ARRAY_FLOAT
        Parameters of the truncated normal distribution.
    lower_bound : float
        Lower bound of the truncated normal distribution.
    upper_bound : float
        Upper bound of the truncated normal distribution.

    Returns
    -------
    ARRAY_FLOAT
        CDF values of the truncated normal distribution on the sample values.

    Notes
    -----
    - CDF for sample with values smaller (resp. larger) than the lower bound
      (resp. upper bound) are set to 0.0 (resp. 1.0).
    - ``lower_bound`` and ``upper_bound`` in the function parameters may not
      be the same as the ones in ``parameters`` as the former are already
      processed setting the numerical bounds when the given bounds are outside
      them.
    """
    yy = np.empty(xx.shape)
    idx_lower = xx < lower_bound
    idx_upper = xx > upper_bound
    idx_rest = np.logical_and(
        np.logical_not(idx_lower), np.logical_not(idx_upper)
    )

    # NOTE: Don't use the original bounds as "lower_bound" and "upper_bound"
    # may have been further truncated for numerical reason.
    mu, sigma, _, _ = _get_parameters(parameters)

    yy[idx_lower] = 0.0
    yy[idx_upper] = 1.0
    yy[idx_rest] = truncnorm.cdf(
        xx[idx_rest],
        a=(lower_bound - mu) / sigma,
        b=(upper_bound - mu) / sigma,
        loc=mu,
        scale=sigma,
    )

    return yy


def icdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the inverse CDF values of a truncated normal distribution.

    Parameters
    ----------
    xx : np.ndarray
        Sample values (realizations) in the [0, 1] domain.
    parameters : np.ndarray
        Parameters of a truncated normal distribution.
    lower_bound : float
        Lower bound of the truncated normal distribution.
    upper_bound : float
        Upper bound of the truncated normal distribution.

    Returns
    -------
    ARRAY_FLOAT
        Transformed values in the domain of the truncated normal distribution.

    Notes
    -----
    - ICDF for sample values outside [0.0, 1.0] is set to NaN according to
      the underlying SciPy implementation.
    - ``lower_bound`` and ``upper_bound`` in the function parameters may not
      be the same as the ones in ``parameters`` as the former are already
      processed setting the numerical bounds when the given bounds are outside
      them.
    """
    # NOTE: Don't use the original bounds as "lower_bound" and "upper_bound"
    # may have been further truncated for numerical reason.
    mu, sigma, _, _ = _get_parameters(parameters)

    # Compute the ICDF
    yy = truncnorm.ppf(
        xx,
        a=(lower_bound - mu) / sigma,
        b=(upper_bound - mu) / sigma,
        loc=mu,
        scale=sigma,
    )

    # Check if values are within the set bounds
    yy = postprocess_icdf(yy, lower_bound, upper_bound)

    return yy



"""
Utility module for probabilistic input modeling.
"""

from numpy.typing import ArrayLike
from types import ModuleType
from typing import Tuple

from .univariate_distributions import (
    beta,
    exponential,
    normal,
    gumbel,
    lognormal,
    triangular,
    trunc_gumbel,
    trunc_normal,
    logitnormal,
    uniform,
)
from ...global_settings import ARRAY_FLOAT

SUPPORTED_MARGINALS = {
    beta.DISTRIBUTION_NAME: beta,
    exponential.DISTRIBUTION_NAME: exponential,
    lognormal.DISTRIBUTION_NAME: lognormal,
    normal.DISTRIBUTION_NAME: normal,
    gumbel.DISTRIBUTION_NAME: gumbel,
    triangular.DISTRIBUTION_NAME: triangular,
    trunc_gumbel.DISTRIBUTION_NAME: trunc_gumbel,
    trunc_normal.DISTRIBUTION_NAME: trunc_normal,
    logitnormal.DISTRIBUTION_NAME: logitnormal,
    uniform.DISTRIBUTION_NAME: uniform,
}


def get_distribution_module(
    distribution: str,
) -> ModuleType:  # pragma: no cover
    """Get the relevant module corresponding to the distribution."""

    distribution_module = SUPPORTED_MARGINALS[distribution]

    return distribution_module


def verify_distribution(distribution: str) -> None:
    """Verify the type of distribution.

    Parameters
    ----------
    distribution : str
        Type of the distribution.

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If the type of distribution is not currently supported.
    """
    if distribution not in SUPPORTED_MARGINALS:
        raise ValueError(
            f"Univariate distribution {distribution!r} is not supported!"
        )


def get_distribution_bounds(
    distribution: str, parameters: ArrayLike
) -> Tuple[float, float]:
    """Get the bounds of the distribution given the parameters.

    While the support of many continuous density functions are unbounded,
    numerically they are always bounded. Below and above the bounds the density
    values are always zero.
    """
    distribution_module = get_distribution_module(distribution)
    lower = distribution_module.lower(parameters)
    upper = distribution_module.upper(parameters)

    return lower, upper


def verify_parameters(distribution: str, parameters: ArrayLike) -> None:
    """Verify the parameter values of the distribution"""
    distribution_module = get_distribution_module(distribution)

    distribution_module.verify_parameters(parameters)


def get_pdf_values(
    xx: ArrayLike,
    distribution: str,
    parameters: ArrayLike,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the PDF values of the distribution on a set of sample points.

    Notes
    -----
    - PDF stands for "probability density function".
    """
    distribution_module = get_distribution_module(distribution)
    out: ARRAY_FLOAT = distribution_module.pdf(
        xx, parameters, lower_bound, upper_bound
    )

    return out


def get_cdf_values(
    xx: ArrayLike,
    distribution: str,
    parameters: ArrayLike,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the CDF values of the distribution on a set of sample points.

    Notes
    -----
    - CDF stands for "cumulative distribution function".
    """
    distribution_module = get_distribution_module(distribution)
    out: ARRAY_FLOAT = distribution_module.cdf(
        xx, parameters, lower_bound, upper_bound
    )

    return out


def get_icdf_values(
    xx: ArrayLike,
    distribution: str,
    parameters: ArrayLike,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the inverse CDF values of the dist. on a set of sample points.

    Notes
    -----
    - CDF stands for "cumulative distribution function".
    - ICDF stands for "inverse cumulative distribution function".
    """
    distribution_module = get_distribution_module(distribution)
    out: ARRAY_FLOAT = distribution_module.icdf(
        xx, parameters, lower_bound, upper_bound
    )

    return out
